@@ -182,7 +182,7 @@ public:
       // prepare for result
       d_resnum = mysql_stmt_num_rows(d_stmt);
       
-      if (d_resnum>0 && d_res_bind == NULL) {
+      if (d_resnum > 0 && d_res_bind == nullptr) {
         MYSQL_RES* meta = mysql_stmt_result_metadata(d_stmt);
         d_fnum = static_cast<int>(mysql_num_fields(meta)); // ensure correct number of fields
         d_res_bind = new MYSQL_BIND[d_fnum];
@@ -201,12 +201,17 @@ public:
         }
   
         mysql_free_result(meta);
-  
-        if ((err = mysql_stmt_bind_result(d_stmt, d_res_bind))) {
-          string error(mysql_stmt_error(d_stmt));
-          releaseStatement();
-          throw SSqlException("Could not bind parameters to mysql statement: " + d_query + string(": ") + error);
-        }
+      }
+
+      /* we need to bind the results array again because a call to mysql_stmt_next_result() followed
+         by a call to mysql_stmt_store_result() might have invalidated it (the first one sets
+         stmt->bind_result_done to false, causing the second to reset the existing binding),
+         and we can't bind it right after the call to mysql_stmt_store_result() if it returned
+         no rows, because then the statement 'contains no metadata' */
+      if (d_res_bind != nullptr && (err = mysql_stmt_bind_result(d_stmt, d_res_bind))) {
+        string error(mysql_stmt_error(d_stmt));
+        releaseStatement();
+        throw SSqlException("Could not bind parameters to mysql statement: " + d_query + string(": ") + error);
       }
     }
 
@@ -220,9 +225,11 @@ public:
   SSqlStatement* nextRow(row_t& row) {
     int err;
     row.clear();
-    if (!hasNextRow()) return this;
+    if (!hasNextRow()) {
+      return this;
+    }
 
-    if ((err =mysql_stmt_fetch(d_stmt))) {
+    if ((err = mysql_stmt_fetch(d_stmt))) {
       if (err != MYSQL_DATA_TRUNCATED) {
         string error(mysql_stmt_error(d_stmt));
         releaseStatement();
@@ -233,8 +240,8 @@ public:
     row.reserve(d_fnum);
 
     for(int i=0;i<d_fnum;i++) {
-      if (*d_res_bind[i].error) {
-        L<<Logger::Warning<<"Result field at row " << d_residx << " column " << i << " has errno " << *d_res_bind[i].error << endl;
+      if (err == MYSQL_DATA_TRUNCATED && *d_res_bind[i].error) {
+        L<<Logger::Warning<<"Result field at row " << d_residx << " column " << i << " has been truncated, we allocated " << d_res_bind[i].buffer_length << " bytes but at least " << *d_res_bind[i].length << " was needed" << endl;
       }
       if (*d_res_bind[i].is_null) {
         row.push_back("");
@@ -252,13 +259,13 @@ public:
         if ((err = mysql_stmt_store_result(d_stmt))) {
           string error(mysql_stmt_error(d_stmt));
           releaseStatement();
-          throw SSqlException("Could not store mysql statement: " + d_query + string(": ") + error);
+          throw SSqlException("Could not store mysql statement while processing additional sets: " + d_query + string(": ") + error);
         }
         d_resnum = mysql_stmt_num_rows(d_stmt);
         // XXX: For some reason mysql_stmt_result_metadata returns NULL here, so we cannot
         // ensure row field count matches first result set.
-        if (d_resnum>0) { // ignore empty result set
-          if ((err = mysql_stmt_bind_result(d_stmt, d_res_bind))) {
+        if (d_resnum > 0) { // ignore empty result set
+          if (d_res_bind != nullptr && (err = mysql_stmt_bind_result(d_stmt, d_res_bind))) {
             string error(mysql_stmt_error(d_stmt));
             releaseStatement();
             throw SSqlException("Could not bind parameters to mysql statement: " + d_query + string(": ") + error);
@@ -393,8 +400,7 @@ private:
   int d_residx;
 };
 
-SMySQL::SMySQL(const string &database, const string &host, uint16_t port, const string &msocket, const string &user,
-               const string &password, const string &group, bool setIsolation, unsigned int timeout)
+void SMySQL::connect()
 {
   int retry=1;
 
@@ -410,9 +416,9 @@ SMySQL::SMySQL(const string &database, const string &host, uint16_t port, const
 #endif
 
 #if MYSQL_VERSION_ID >= 50100
-    if(timeout) {
-      mysql_options(&d_db, MYSQL_OPT_READ_TIMEOUT, &timeout);
-      mysql_options(&d_db, MYSQL_OPT_WRITE_TIMEOUT, &timeout);
+    if(d_timeout) {
+      mysql_options(&d_db, MYSQL_OPT_READ_TIMEOUT, &d_timeout);
+      mysql_options(&d_db, MYSQL_OPT_WRITE_TIMEOUT, &d_timeout);
     }
 #endif
 
@@ -420,18 +426,18 @@ SMySQL::SMySQL(const string &database, const string &host, uint16_t port, const
     mysql_options(&d_db, MYSQL_SET_CHARSET_NAME, MYSQL_AUTODETECT_CHARSET_NAME);
 #endif
 
-    if (setIsolation && (retry == 1))
+    if (d_setIsolation && (retry == 1))
       mysql_options(&d_db, MYSQL_INIT_COMMAND,"SET SESSION tx_isolation='READ-COMMITTED'");
 
-    mysql_options(&d_db, MYSQL_READ_DEFAULT_GROUP, group.c_str());
+    mysql_options(&d_db, MYSQL_READ_DEFAULT_GROUP, d_group.c_str());
 
-    if (!mysql_real_connect(&d_db, host.empty() ? NULL : host.c_str(),
-                          user.empty() ? NULL : user.c_str(),
-                          password.empty() ? NULL : password.c_str(),
-                          database.empty() ? NULL : database.c_str(),
-                          port,
-                          msocket.empty() ? NULL : msocket.c_str(),
-                          CLIENT_MULTI_RESULTS)) {
+    if (!mysql_real_connect(&d_db, d_host.empty() ? NULL : d_host.c_str(),
+                            d_user.empty() ? NULL : d_user.c_str(),
+                            d_password.empty() ? NULL : d_password.c_str(),
+                            d_database.empty() ? NULL : d_database.c_str(),
+                            d_port,
+                            d_msocket.empty() ? NULL : d_msocket.c_str(),
+                            CLIENT_MULTI_RESULTS)) {
 
       if (retry == 0)
         throw sPerrorException("Unable to connect to database");
@@ -446,6 +452,13 @@ SMySQL::SMySQL(const string &database, const string &host, uint16_t port, const
   } while (retry >= 0);
 }
 
+SMySQL::SMySQL(const string &database, const string &host, uint16_t port, const string &msocket, const string &user,
+               const string &password, const string &group, bool setIsolation, unsigned int timeout):
+  d_database(database), d_host(host), d_msocket(msocket), d_user(user), d_password(password), d_group(group), d_timeout(timeout), d_port(port), d_setIsolation(setIsolation)
+{
+  connect();
+}
+
 void SMySQL::setLog(bool state)
 {
   s_dolog=state;
@@ -461,9 +474,9 @@ SSqlException SMySQL::sPerrorException(const string &reason)
   return SSqlException(reason+string(": ")+mysql_error(&d_db));
 }
 
-SSqlStatement* SMySQL::prepare(const string& query, int nparams)
+std::unique_ptr<SSqlStatement> SMySQL::prepare(const string& query, int nparams)
 {
-  return new SMySQLStatement(query, s_dolog, nparams, &d_db);
+  return std::unique_ptr<SSqlStatement>(new SMySQLStatement(query, s_dolog, nparams, &d_db));
 }
 
 void SMySQL::execute(const string& query)
@@ -487,3 +500,26 @@ void SMySQL::commit() {
 void SMySQL::rollback() {
   execute("rollback");
 }
+
+bool SMySQL::isConnectionUsable()
+{
+  bool usable = false;
+  int sd = d_db.net.fd;
+  bool wasNonBlocking = isNonBlocking(sd);
+
+  if (!wasNonBlocking) {
+    if (!setNonBlocking(sd)) {
+      return usable;
+    }
+  }
+
+  usable = isTCPSocketUsable(sd);
+
+  if (!wasNonBlocking) {
+    if (!setBlocking(sd)) {
+      usable = false;
+    }
+  }
+
+  return usable;
+}
