@@ -176,12 +176,10 @@ bool LUABackend::get(DNSResourceRecord &rr) {
     return !rr.content.empty();
 }
 
-bool LUABackend::getSOA(const DNSName &name, SOAData &soadata, DNSPacket *p) {
+bool LUABackend::getSOA(const DNSName &name, SOAData &soadata) {
     if (logging)
 	L << Logger::Info << backend_name << "(getsoa) BEGIN" << endl;
 
-    dnspacket = p;
-
     lua_rawgeti(lua, LUA_REGISTRYINDEX, f_lua_getsoa);
 
     lua_pushstring(lua, name.toString().c_str());
@@ -190,14 +188,10 @@ bool LUABackend::getSOA(const DNSName &name, SOAData &soadata, DNSPacket *p) {
 	string e = backend_name + lua_tostring(lua, -1);
 	lua_pop(lua, 1);
 
-	dnspacket = NULL;
-
 	throw runtime_error(e);
 	return false;
     }
 
-    dnspacket = NULL;
-
     size_t returnedwhat = lua_type(lua, -1);
     if (returnedwhat != LUA_TTABLE) {
 	lua_pop(lua, 1 );
@@ -206,6 +200,7 @@ bool LUABackend::getSOA(const DNSName &name, SOAData &soadata, DNSPacket *p) {
 
     soadata.db = this;
     soadata.serial = 0;
+    soadata.qname = name;
     getValueFromTable(lua, "serial", soadata.serial);
     if (soadata.serial == 0) {
 	lua_pop(lua, 1 );
