@@ -33,13 +33,11 @@
 #include "pdns/dns.hh"
 #include "pdns/dnsbackend.hh"
 #include "pdns/dnspacket.hh"
-#include "pdns/ueberbackend.hh"
 #include "pdns/pdnsexception.hh"
 #include "pdns/logger.hh"
 #include "pdns/arguments.hh"
 #include "pdns/json.hh"
 #include "pdns/statbag.hh"
-#include "pdns/packetcache.hh"
 
 #include "test-remotebackend-keys.hh"
 
@@ -115,15 +113,18 @@ BOOST_AUTO_TEST_CASE(test_method_getAllDomainMetadata) {
 
 BOOST_AUTO_TEST_CASE(test_method_addDomainKey) {
    BOOST_TEST_MESSAGE("Testing addDomainKey method");
-   BOOST_CHECK_EQUAL(be->addDomainKey(DNSName("unit.test."),k1), 1);
-   BOOST_CHECK_EQUAL(be->addDomainKey(DNSName("unit.test."),k2), 2);
+   int64_t id;
+   be->addDomainKey(DNSName("unit.test."),k1,id);
+   BOOST_CHECK_EQUAL(id, 1);
+   be->addDomainKey(DNSName("unit.test."),k2,id);
+   BOOST_CHECK_EQUAL(id, 2);
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getDomainKeys) {
    std::vector<DNSBackend::KeyData> keys;
    BOOST_TEST_MESSAGE("Testing getDomainKeys method");
    // we expect to get two keys
-   be->getDomainKeys(DNSName("unit.test."),0,keys);
+   be->getDomainKeys(DNSName("unit.test."),keys);
    BOOST_CHECK_EQUAL(keys.size(), 2);
    // in case we got more than 2 keys, which would be unexpected
    // but not fatal
@@ -154,14 +155,13 @@ BOOST_AUTO_TEST_CASE(test_method_removeDomainKey) {
 }
 
 BOOST_AUTO_TEST_CASE(test_method_getBeforeAndAfterNamesAbsolute) {
-   DNSName unhashed;
-   std::string before,after;
+   DNSName unhashed, before, after;
    BOOST_TEST_MESSAGE("Testing getBeforeAndAfterNamesAbsolute method");
    
-   be->getBeforeAndAfterNamesAbsolute(-1, "middle.unit.test.", unhashed, before, after);
+   be->getBeforeAndAfterNamesAbsolute(-1, DNSName("middle.unit.test."), unhashed, before, after);
    BOOST_CHECK_EQUAL(unhashed.toString(), "middle.");
-   BOOST_CHECK_EQUAL(before, "begin.");
-   BOOST_CHECK_EQUAL(after, "stop.");
+   BOOST_CHECK_EQUAL(before.toString(), "begin.");
+   BOOST_CHECK_EQUAL(after.toString(), "stop.");
 }
 
 BOOST_AUTO_TEST_CASE(test_method_setTSIGKey) {
@@ -274,13 +274,13 @@ BOOST_AUTO_TEST_CASE(test_method_feedRecord) {
    rr.qclass = QClass::IN;
    rr.ttl = 300;
    rr.content = "ns1.example.com. hostmaster.example.com. 2013013441 7200 3600 1209600 300";
-   BOOST_CHECK(be->feedRecord(rr, NULL));
+   BOOST_CHECK(be->feedRecord(rr, DNSName()));
    rr.qname = DNSName("replace.example.com.");
    rr.qtype = QType::A;
    rr.qclass = QClass::IN;
    rr.ttl = 300;
    rr.content = "127.0.0.1";
-   BOOST_CHECK(be->feedRecord(rr, NULL));
+   BOOST_CHECK(be->feedRecord(rr, DNSName()));
    be->commitTransaction();
 }
 
