@@ -22,22 +22,25 @@
 #ifdef HAVE_CONFIG_H
 #include "config.h"
 #endif
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/un.h>
 #include "remotebackend.hh"
 #ifndef UNIX_PATH_MAX 
 #define UNIX_PATH_MAX 108
 #endif
 
-UnixsocketConnector::UnixsocketConnector(std::map<std::string,std::string> options) {
-  if (options.count("path") == 0) {
+UnixsocketConnector::UnixsocketConnector(std::map<std::string,std::string> optionsMap) {
+  if (optionsMap.count("path") == 0) {
     L<<Logger::Error<<"Cannot find 'path' option in connection string"<<endl;
     throw PDNSException();
   } 
   this->timeout = 2000;
-  if (options.find("timeout") != options.end()) { 
-    this->timeout = std::stoi(options.find("timeout")->second);
+  if (optionsMap.find("timeout") != optionsMap.end()) {
+    this->timeout = std::stoi(optionsMap.find("timeout")->second);
   }
-  this->path = options.find("path")->second;
-  this->options = options;
+  this->path = optionsMap.find("path")->second;
+  this->options = optionsMap;
   this->connected = false;
   this->fd = -1;
 }
@@ -62,12 +65,11 @@ int UnixsocketConnector::send_message(const Json& input) {
 }
 
 int UnixsocketConnector::recv_message(Json& output) {
-  int rv,nread;
+  int rv;
   std::string s_output,err;
 
   struct timeval t0,t;
 
-  nread = 0;
   gettimeofday(&t0, NULL);
   memcpy(&t,&t0,sizeof(t0));
   s_output = "";       
@@ -81,16 +83,11 @@ int UnixsocketConnector::recv_message(Json& output) {
       continue;
     }
 
-    std::string temp;
-    temp.clear();
-
-    rv = this->read(temp);
+    rv = this->read(s_output);
     if (rv == -1) 
       return -1;
 
     if (rv>0) {
-      nread += rv;
-      s_output.append(temp);
       // see if it can be parsed
       output = Json::parse(s_output, err);
       if (output != nullptr) return s_output.size();
@@ -125,25 +122,22 @@ ssize_t UnixsocketConnector::read(std::string &data) {
 }
 
 ssize_t UnixsocketConnector::write(const std::string &data) {
-  ssize_t nwrite, nbuf;
-  size_t pos;
-  char buf[1500];
+  size_t pos = 0;
 
   reconnect();
   if (!connected) return -1;
-  pos = 0;
-  nwrite = 0;
+
   while(pos < data.size()) {
-    nbuf = data.copy(buf, sizeof buf, pos); // copy data and write
-    nwrite = ::write(fd, buf, nbuf);
-    pos = pos + sizeof(buf);
-    if (nwrite < 1) {
+    ssize_t written = ::write(fd, &data.at(pos), data.size() - pos);
+    if (written < 1) {
       connected = false;
       close(fd);
       return -1;
+    } else {
+      pos = pos + static_cast<size_t>(written);
     }
   }
-  return nwrite;
+  return pos;
 }
 
 void UnixsocketConnector::reconnect() {
