@@ -276,7 +276,7 @@ bool TinyDNSBackend::get(DNSResourceRecord &rr)
       rr.qname=DNSName(key.c_str(), key.size(), 0, false);
       rr.domain_id=-1;
       // 11:13.21 <@ahu> IT IS ALWAYS AUTH --- well not really because we are just a backend :-)
-      // We could actually do NSEC3-NARROW DNSSEC according to Habbie, if we do, we need to change something ehre.
+      // We could actually do NSEC3-NARROW DNSSEC according to Habbie, if we do, we need to change something here.
       rr.auth = true;
 
       rr.ttl = pr.get32BitInt();
@@ -300,17 +300,19 @@ bool TinyDNSBackend::get(DNSResourceRecord &rr)
         dr.d_type = rr.qtype.getCode();
         dr.d_clen = val.size()-pr.d_pos;
 
-        DNSRecordContent *drc = DNSRecordContent::mastermake(dr, pr);
+        auto drc = DNSRecordContent::mastermake(dr, pr);
         rr.content = drc->getZoneRepresentation();
         DLOG(cerr<<"CONTENT: "<<rr.content<<endl);
-        delete drc;
       }
       catch (...) {
+        L<<Logger::Error<<backendname<<"Failed to parse record content for "<<rr.qname<<" with type "<<rr.qtype.getName();
         if (d_ignorebogus) {
-          L<<Logger::Error<<backendname<<"Failed to parse record content for "<<rr.qname<<" with type "<<rr.qtype.getName()<<". Ignoring!"<<endl;
+          L<<". Ignoring!"<<endl;
           continue;
-        } else
+        } else {
+          L<<". Erroring out!"<<endl;
           throw;
+        }
       }
 //      DLOG(L<<Logger::Debug<<backendname<<"Returning ["<<rr.content<<"] for ["<<rr.qname<<"] of RecordType ["<<rr.qtype.getName()<<"]"<<endl;);
       return true;
