@@ -55,6 +55,22 @@ void CommunicatorClass::retrievalLoopThread(void)
   }
 }
 
+void CommunicatorClass::loadArgsIntoSet(const char *listname, set<string> &listset)
+{
+  vector<string> parts;
+  stringtok(parts, ::arg()[listname], ", \t");
+  for (vector<string>::const_iterator iter = parts.begin(); iter != parts.end(); ++iter) {
+    try {
+      ComboAddress caIp(*iter, 53);
+      listset.insert(caIp.toStringWithPort());
+    }
+    catch(PDNSException &e) {
+      L<<Logger::Error<<"Unparseable IP in "<<listname<<". Error: "<<e.reason<<endl;
+      _exit(1);
+    }
+  }
+}
+
 void CommunicatorClass::go()
 {
   try {
@@ -62,7 +78,7 @@ void CommunicatorClass::go()
   }
   catch(PDNSException &e) {
     L<<Logger::Error<<"Unparseable IP in allow-notify-from. Error: "<<e.reason<<endl;
-    exit(1);
+    _exit(1);
   }
 
   pthread_t tid;
@@ -77,21 +93,12 @@ void CommunicatorClass::go()
   }
   catch(PDNSException &e) {
     L<<Logger::Error<<"Unparseable IP in only-notify. Error: "<<e.reason<<endl;
-    exit(1);
+    _exit(1);
   }
 
-  vector<string> parts;
-  stringtok(parts, ::arg()["also-notify"], ", \t");
-  for (vector<string>::const_iterator iter = parts.begin(); iter != parts.end(); ++iter) {
-    try {
-      ComboAddress caIp(*iter, 53);
-      d_alsoNotify.insert(caIp.toStringWithPort());
-    }
-    catch(PDNSException &e) {
-      L<<Logger::Error<<"Unparseable IP in also-notify. Error: "<<e.reason<<endl;
-      exit(1);
-    }
-  }
+  loadArgsIntoSet("also-notify", d_alsoNotify);
+
+  loadArgsIntoSet("forward-notify", PacketHandler::s_forwardNotify);
 }
 
 void CommunicatorClass::mainloop(void)
@@ -131,16 +138,16 @@ void CommunicatorClass::mainloop(void)
   catch(PDNSException &ae) {
     L<<Logger::Error<<"Exiting because communicator thread died with error: "<<ae.reason<<endl;
     Utility::sleep(1);
-    exit(1);
+    _exit(1);
   }
   catch(std::exception &e) {
     L<<Logger::Error<<"Exiting because communicator thread died with STL error: "<<e.what()<<endl;
-    exit(1);
+    _exit(1);
   }
   catch( ... )
   {
     L << Logger::Error << "Exiting because communicator caught unknown exception." << endl;
-    exit(1);
+    _exit(1);
   }
 }
 
