@@ -25,7 +25,7 @@ Replay all recursion-desired DNS questions to a specified IP address.
 Track all outgoing questions, remap id to one of ours.
 Also track all recorded answers, and map them to that same id, the 'expectation'.
 
-When we see a question, parse it, give it a QuestionIdentifyer, and and an id from the free-id list.
+When we see a question, parse it, give it a QuestionIdentifier, and and an id from the free-id list.
 
 When we see an answer in the tcpdump, parse it, make QI, and add it to the original QI
    and check
@@ -649,17 +649,17 @@ bool sendPacketFromPR(PcapPacketReader& pr, const ComboAddress& remote, int stam
       s_origanswers++;
       qids_t::const_iterator iter=qids.find(qi);      
       if(iter != qids.end()) {
-        QuestionData qd=*iter;
-        qd.d_origAnswers=mdp.d_answers;
-        qd.d_origRcode=mdp.d_header.rcode;
+        QuestionData eqd=*iter;
+        eqd.d_origAnswers=mdp.d_answers;
+        eqd.d_origRcode=mdp.d_header.rcode;
         
         if(!dh->ra) {
           s_norecursionavailable++;
-          qd.d_norecursionavailable=true;
+          eqd.d_norecursionavailable=true;
         }
-        qids.replace(iter, qd);
+        qids.replace(iter, eqd);
 
-        if(qd.d_newRcode!=-1) {
+        if(eqd.d_newRcode!=-1) {
           measureResultAndClean(iter);
         }
         
@@ -808,7 +808,7 @@ try
       if(sendPacketFromPR(pr, remote, stamp))
         count++;
     } 
-    if(packetLimit && count > packetLimit) 
+    if(packetLimit && count >= packetLimit) 
       break;
 
     mental_time=packet_ts;
