@@ -1,21 +1,44 @@
 
+#include "gettime.hh"
+#include "dnsparser.hh"
 #include "protobuf.hh"
 #include "dnsparser.hh"
 #include "gettime.hh"
 
-DNSProtoBufMessage::DNSProtoBufMessage(DNSProtoBufMessageType type)
+void DNSProtoBufMessage::setType(DNSProtoBufMessageType type)
 {
 #ifdef HAVE_PROTOBUF
-  d_message.set_type(type == DNSProtoBufMessage::DNSProtoBufMessageType::Query ? PBDNSMessage_Type_DNSQueryType : PBDNSMessage_Type_DNSResponseType);
+  switch(type) {
+  case DNSProtoBufMessage::DNSProtoBufMessageType::Query:
+    d_message.set_type(PBDNSMessage_Type_DNSQueryType);
+    break;
+  case DNSProtoBufMessage::DNSProtoBufMessageType::Response:
+    d_message.set_type(PBDNSMessage_Type_DNSResponseType);
+    break;
+  case DNSProtoBufMessage::DNSProtoBufMessageType::OutgoingQuery:
+    d_message.set_type(PBDNSMessage_Type_DNSOutgoingQueryType);
+    break;
+  case DNSProtoBufMessage::DNSProtoBufMessageType::IncomingResponse:
+    d_message.set_type(PBDNSMessage_Type_DNSIncomingResponseType);
+    break;
+  default:
+    throw std::runtime_error("Unsupported protobuf type: "+std::to_string(type));
+  }
 #endif /* HAVE_PROTOBUF */
 }
 
+DNSProtoBufMessage::DNSProtoBufMessage(DNSProtoBufMessageType type)
+{
+  setType(type);
+}
+
 void DNSProtoBufMessage::setQuestion(const DNSName& qname, uint16_t qtype, uint16_t qclass)
 {
 #ifdef HAVE_PROTOBUF
   PBDNSMessage_DNSQuestion* question = d_message.mutable_question();
   if (question) {
-    question->set_qname(qname.toString());
+    if(!qname.empty())
+      question->set_qname(qname.toString());
     question->set_qtype(qtype);
     question->set_qclass(qclass);
   }
@@ -74,7 +97,39 @@ void DNSProtoBufMessage::setEDNSSubnet(const Netmask& subnet, uint8_t mask)
 #endif /* HAVE_PROTOBUF */
 }
 
-void DNSProtoBufMessage::addRRsFromPacket(const char* packet, const size_t len)
+void DNSProtoBufMessage::addTag(const std::string& strValue)
+{
+#ifdef HAVE_PROTOBUF
+
+  PBDNSMessage_DNSResponse* response = d_message.mutable_response();
+  if (!response)
+    return;
+
+  response->add_tags(strValue);
+
+#endif /* HAVE_PROTOBUF */
+}
+
+void DNSProtoBufMessage::addRR(const DNSName& qname, uint16_t uType, uint16_t uClass, uint32_t uTTL, const std::string& strBlob)
+{
+#ifdef HAVE_PROTOBUF
+
+  PBDNSMessage_DNSResponse* response = d_message.mutable_response();
+  if (!response)
+    return;
+  PBDNSMessage_DNSResponse_DNSRR* rr = response->add_rrs();
+  if (rr) {
+    rr->set_name(qname.toString());
+    rr->set_type(uType);
+    rr->set_class_(uClass);
+    rr->set_ttl(uTTL);
+    rr->set_rdata(strBlob.c_str(), strBlob.size());
+  }
+
+#endif /* HAVE_PROTOBUF */
+}
+
+void DNSProtoBufMessage::addRRsFromPacket(const char* packet, const size_t len, bool includeCNAME)
 {
 #ifdef HAVE_PROTOBUF
   if (len < sizeof(struct dnsheader))
@@ -125,9 +180,9 @@ void DNSProtoBufMessage::addRRsFromPacket(const char* packet, const size_t len)
     rrname = pr.getName();
     pr.getDnsrecordheader(ah);
 
-    pr.xfrBlob(blob);
-
     if (ah.d_type == QType::A || ah.d_type == QType::AAAA) {
+      pr.xfrBlob(blob);
+
       PBDNSMessage_DNSResponse_DNSRR* rr = response->add_rrs();
       if (rr) {
         rr->set_name(rrname.toString());
@@ -136,6 +191,20 @@ void DNSProtoBufMessage::addRRsFromPacket(const char* packet, const size_t len)
         rr->set_ttl(ah.d_ttl);
         rr->set_rdata(blob.c_str(), blob.length());
       }
+    } else if (ah.d_type == QType::CNAME && includeCNAME) {
+      PBDNSMessage_DNSResponse_DNSRR* rr = response->add_rrs();
+      if (rr) {
+        rr->set_name(rrname.toString());
+        rr->set_type(ah.d_type);
+        rr->set_class_(ah.d_class);
+        rr->set_ttl(ah.d_ttl);
+        DNSName target;
+        pr.xfrName(target, true);
+        rr->set_rdata(target.toString());
+      }
+    }
+    else {
+      pr.xfrBlob(blob);
     }
   }
 #endif /* HAVE_PROTOBUF */
@@ -160,10 +229,24 @@ void DNSProtoBufMessage::setRequestor(const ComboAddress& requestor)
 #endif /* HAVE_PROTOBUF */
 }
 
+void DNSProtoBufMessage::setRequestorId(const std::string& requestorId)
+{
+#ifdef HAVE_PROTOBUF
+  d_message.set_requestorid(requestorId);
+#endif /* HAVE_PROTOBUF */
+}
+
+void DNSProtoBufMessage::setDeviceId(const std::string& deviceId)
+{
+#ifdef HAVE_PROTOBUF
+  d_message.set_deviceid(deviceId);
+#endif /* HAVE_PROTOBUF */
+}
+
 void DNSProtoBufMessage::setResponder(const std::string& responder)
 {
 #ifdef HAVE_PROTOBUF
-  d_message.set_from(responder);
+  d_message.set_to(responder);
 #endif /* HAVE_PROTOBUF */
 }
 
@@ -171,10 +254,10 @@ void DNSProtoBufMessage::setResponder(const ComboAddress& responder)
 {
 #ifdef HAVE_PROTOBUF
   if (responder.sin4.sin_family == AF_INET) {
-    d_message.set_from(&responder.sin4.sin_addr.s_addr, sizeof(responder.sin4.sin_addr.s_addr));
+    d_message.set_to(&responder.sin4.sin_addr.s_addr, sizeof(responder.sin4.sin_addr.s_addr));
   }
   else if (responder.sin4.sin_family == AF_INET6) {
-    d_message.set_from(&responder.sin6.sin6_addr.s6_addr, sizeof(responder.sin6.sin6_addr.s6_addr));
+    d_message.set_to(&responder.sin6.sin6_addr.s6_addr, sizeof(responder.sin6.sin6_addr.s6_addr));
   }
 #endif /* HAVE_PROTOBUF */
 }
@@ -204,6 +287,13 @@ void DNSProtoBufMessage::setUUID(const boost::uuids::uuid& uuid)
   std::copy(uuid.begin(), uuid.end(), messageId->begin());
 }
 
+void DNSProtoBufMessage::setInitialRequestID(const boost::uuids::uuid& uuid)
+{
+  std::string* messageId = d_message.mutable_initialrequestid();
+  messageId->resize(uuid.size());
+  std::copy(uuid.begin(), uuid.end(), messageId->begin());
+}
+
 void DNSProtoBufMessage::update(const boost::uuids::uuid& uuid, const ComboAddress* requestor, const ComboAddress* responder, bool isTCP, uint16_t id)
 {
   struct timespec ts;
@@ -213,7 +303,13 @@ void DNSProtoBufMessage::update(const boost::uuids::uuid& uuid, const ComboAddre
   setUUID(uuid);
   d_message.set_id(ntohs(id));
 
-  d_message.set_socketfamily((requestor && requestor->sin4.sin_family == AF_INET) ? PBDNSMessage_SocketFamily_INET : PBDNSMessage_SocketFamily_INET6);
+  if (requestor) {
+    d_message.set_socketfamily(requestor->sin4.sin_family == AF_INET ? PBDNSMessage_SocketFamily_INET : PBDNSMessage_SocketFamily_INET6);
+  }
+  else if (responder) {
+    d_message.set_socketfamily(responder->sin4.sin_family == AF_INET ? PBDNSMessage_SocketFamily_INET : PBDNSMessage_SocketFamily_INET6);
+  }
+
   d_message.set_socketprotocol(isTCP ? PBDNSMessage_SocketProtocol_TCP : PBDNSMessage_SocketProtocol_UDP);
 
   if (responder) {
@@ -225,11 +321,11 @@ void DNSProtoBufMessage::update(const boost::uuids::uuid& uuid, const ComboAddre
 }
 
 
-DNSProtoBufMessage::DNSProtoBufMessage(DNSProtoBufMessageType type, const boost::uuids::uuid& uuid, const ComboAddress* requestor, const ComboAddress* to, const DNSName& domain, int qtype, uint16_t qclass, uint16_t qid, bool isTCP, size_t bytes)
+DNSProtoBufMessage::DNSProtoBufMessage(DNSProtoBufMessageType type, const boost::uuids::uuid& uuid, const ComboAddress* requestor, const ComboAddress* responder, const DNSName& domain, int qtype, uint16_t qclass, uint16_t qid, bool isTCP, size_t bytes)
 {
-  update(uuid, requestor, to, isTCP, qid);
+  update(uuid, requestor, responder, isTCP, qid);
 
-  d_message.set_type(type == DNSProtoBufMessage::DNSProtoBufMessageType::Query ? PBDNSMessage_Type_DNSQueryType : PBDNSMessage_Type_DNSResponseType);
+  setType(type);
 
   setBytes(bytes);
   setQuestion(domain, qtype, qclass);
