@@ -15,6 +15,7 @@
 #include "namespaces.hh"
 #include "statbag.hh"
 #include "stubresolver.hh"
+#include "dnsrecords.hh"
 #include <stdint.h>
 #ifndef PACKAGEVERSION
 #define PACKAGEVERSION getPDNSVersion()
@@ -44,19 +45,16 @@ void doSecPoll(bool first)
   boost::replace_all(query, "+", "_");
   boost::replace_all(query, "~", "_");
 
-  vector<DNSResourceRecord> ret;
+  vector<DNSZoneRecord> ret;
 
-  int res=stubDoResolve(query, QType::TXT, ret);
+  int res=stubDoResolve(DNSName(query), QType::TXT, ret);
 
   int security_status=0;
 
   if(!res && !ret.empty()) {
-    string content=ret.begin()->content;
-    if(!content.empty() && content[0]=='"' && content[content.size()-1]=='"') {
-      content=content.substr(1, content.length()-2);
-    }
+    string content=getRR<TXTRecordContent>(ret.begin()->dr)->d_text;
 
-    pair<string, string> split = splitField(content, ' ');
+    pair<string, string> split = splitField(unquotify(content), ' ');
 
     security_status = std::stoi(split.first);
     g_security_message = split.second;
@@ -64,10 +62,10 @@ void doSecPoll(bool first)
   }
   else {
     string pkgv(PACKAGEVERSION);
-    if(pkgv.find("0.0."))
+    if(pkgv.find("0.0.") != 0)
       L<<Logger::Warning<<"Could not retrieve security status update for '" + pkgv + "' on '"+query+"', RCODE = "<< RCode::to_s(res)<<endl;
     else
-      L<<Logger::Warning<<"Not validating response for security status update, this a non-release version."<<endl;
+      L<<Logger::Warning<<"Not validating response for security status update, this is a non-release version."<<endl;
   }
 
   if(security_status == 1 && first) {
